import sys
import signal
from multiprocessing import active_children, Value, Manager, Event
from queue import Empty
from itertools import zip_longest

from cml.domain.deconstruction import *
from cml.domain.data_source import DataSource
from cml.domain.knowledge import KnowledgeDatabase, RelativeFinder, TS_QUEUE
from cml.shared.settings import DeconstructionSettings
from cml.shared.logger import DeconstLoggger, DeconstBanner


__all__ = (
    "Deconstructor",
)

def handle_background_deconstructors_failure(signum, stack):
    sys.stdout.write("An exception occured in one of the deconstructors. "
                     "Tried to deconstructor a model. "
                     "Refer to stderr to identify the exception cause.")
    for child in active_children():
        child.kill()
    sys.exit(-777)
signal.signal(signal.SIGTERM, handle_background_deconstructors_failure)


class ChangeSerializier:
    def __init__(self, complete_change, z_sigma_change, t_sigma_change, pragmatic):
        self.pragmatic = pragmatic
        self.complete = complete_change
        self.z_sigma = z_sigma_change
        self.t_sigma = t_sigma_change

    def serialize(self):
        sorted_changes = self.sort_changes(self.complete, self.z_sigma, self.t_sigma)
        valid_changes = self.correct(len(sorted_changes)-1, sorted_changes)
        return valid_changes

    def correct(self, idx, changes):
        if not changes:
            return changes
        elif idx < 0:
            return changes
        else:
            change = changes[idx]
            delete, repair = self.identify_inconsistencies(change, changes)
            corrected_changes = self.repair_inconsistencies(repair, changes, change)
            corrected_changes = self.delete_inconsistencies(delete, corrected_changes)
            if len(changes) == len(corrected_changes):
                idx -= 1
            elif len(changes) > len(corrected_changes):
                idx = len(corrected_changes)-1
            return self.correct(idx, corrected_changes)

    def identify_inconsistencies(self, from_change, changes):
        to_delete = []
        to_repair = []
        deleted_models = set(from_change.deleted)
        for idx, change in enumerate(changes):
            if self.pragmatic in change.deleted:
                to_delete.extend(range(idx, len(changes)))
                break

            if (from_change is not change) and \
                    {change.relative}.intersection(deleted_models) or \
                    set(change.inserted).intersection(deleted_models):
                to_delete.append(idx)

            if (from_change is not change) and \
                    set(change.deleted).intersection(deleted_models):
                to_repair.append(idx)

        return to_delete, to_repair

    def sort_changes(self, complete_changes, z_sigma_changes, t_sigma_changes):
        complete_changes.sort(key=lambda x: x.time_stamp, reverse=True)
        z_sigma_changes.sort(key=lambda x: x.time_stamp, reverse=True)
        t_sigma_changes.sort(key=lambda x: x.time_stamp, reverse=True)
        return t_sigma_changes + z_sigma_changes + complete_changes

    def delete_inconsistencies(self, to_delete, changes):
        filtered_changes = []
        for idx, change in enumerate(changes):
            if idx not in to_delete:
                filtered_changes.append(change)
        return filtered_changes

    def repair_inconsistencies(self, to_repair, changes, change):
        for idx in to_repair:
            changes[idx].deleted = list(set(changes[idx].deleted).difference(
                set(change.deleted)))
        return changes

    def upgrade_inconsistencies(self, to_upgrade):
        add = 0
        for change in to_upgrade:
            for inserted in change.inserted:
                inserted.identifier += add
                add += 1


class Deconstructor:
    TIME_COLUMN = "T"
    SUBJECT_COLUMN = "Sigma"
    PURPOSE_COLUMN = "Z"

    def __init__(self,
                 mode,
                 knowledge_database: KnowledgeDatabase,
                 relative_finder: RelativeFinder,
                 source: DataSource,
                 settings: DeconstructionSettings):
        self.mode = mode
        self.source = source
        self.settings = settings
        self.relative_finder = relative_finder
        self.knowledge_database = knowledge_database
        self.next_model_counter = knowledge_database.model_counter()
        self.logger = None
        self.reconstructor = None
        self.CHANGES = None

    def args(self, tier, relationship, pragmatic, learnblock):
        relatives = list(self.relative_finder.find(relationship, pragmatic))
        return (tier,
                self.source,
                self.knowledge_database,
                self.settings,
                self.logger,
                self.reconstructor,
                pragmatic,
                relatives,
                learnblock,
                self.CHANGES)

    @DeconstBanner
    def deconstruct_parallel(self, tier, pragmatic, block):
        self.CHANGES = Manager().Queue()
        complete_event = Event()
        t_sigma_event = Event()
        z_sigma_event = Event()

        val = Value("i", self.knowledge_database.model_counter()(tier))

        processes = (
            CompleteDeconstructor(*self.args(tier, ("complete", ), pragmatic, block), val, complete_event),
            TSigmaDeconstructor(*self.args(tier, ("T", "Sigma"), pragmatic, block), val, t_sigma_event),
            ZSigmaDeconstructor(*self.args(tier, ("Sigma", "Z"), pragmatic, block), val, z_sigma_event)
        )

        for process in processes:
            process.daemon = True
            process.start()

        complete_event.wait()
        t_sigma_event.wait()
        z_sigma_event.wait()

        self.collect_results(pragmatic, tier)

    @DeconstBanner
    def deconstruct_serial(self, tier, pragmatic, block):
        val = Value("i", self.knowledge_database.model_counter()(tier))
        self.CHANGES = Manager().Queue()
        processes = (
            CompleteDeconstructor(*self.args(tier, ("complete", ), pragmatic, block), val),
            TSigmaDeconstructor(*self.args(tier, ("T", "Sigma"), pragmatic, block), val),
            ZSigmaDeconstructor(*self.args(tier, ("Sigma", "Z"), pragmatic, block), val)
        )

        for process in processes:
            if process.relatives:
                process.run()

        self.collect_results(pragmatic, tier)

    def collect_results(self, pragmatic, tier):
        changes = self.clear_queue()
        changes_done_status = []
        serializer = ChangeSerializier(*changes, pragmatic)
        for valid_change in serializer.serialize():
            self.apply(valid_change, tier)
            changes_done_status.append(valid_change.status)

            if valid_change.status and self.settings.deconst_mode == "minimal":
                return

        if not changes_done_status or not all(changes_done_status):
            self.logger.protocol("Inserted: {:<20}".format(str(pragmatic)))
            self.knowledge_database.insert(pragmatic)

    @DeconstLoggger
    def apply(self, change, tier):
        if isinstance(change, CompleteChange):
            for d, i in zip_longest(change.deleted, change.inserted):
                if d: self.knowledge_database.remove(d)
                if i: self.knowledge_database.insert(i)

        elif isinstance(change, TSigmaChange) and change.status:
            TS_QUEUE.append((tier+1, change.new_learnblock))

        elif isinstance(change, ZSigmaChange):
            for d, i in zip_longest(change.deleted, change.inserted):
                if d: self.knowledge_database.remove(d)
                if i: self.knowledge_database.insert(i)

    def clear_queue(self):
        complete_changes = []
        z_sigma_changes = []
        t_sigma_changes = []

        while not self.CHANGES.empty():
            try:
                change = self.CHANGES.get(False)
                if isinstance(change, CompleteChange):
                    complete_changes.append(change)
                elif isinstance(change, ZSigmaChange):
                    z_sigma_changes.append(change)
                elif isinstance(change, TSigmaChange):
                    t_sigma_changes.append(change)
            except Empty: continue
        return complete_changes, z_sigma_changes, t_sigma_changes


    # def deconstruct(self,
    #                 tier: int,
    #                 pragmatic: PragmaticMachineLearningModel,
    #                 learnblock) -> None:
    #
    #     success = False
    #     for pair, strategy in self._strategies(learnblock):
    #         for relative in self.relative_finder.find(pair, pragmatic):
    #             try:
    #                 success = strategy(tier, pragmatic, relative, learnblock)
    #                 if success and self.settings.deconst_mode == "minimal":
    #                     return
    #             except NoModelReconstructedError:
    #                 continue
    #             except DeconstructionFailed:
    #                 continue
    #             except Exception as error:
    #                 print(error.with_traceback())
    #
    #     if not success:
    #         # All deconstructions failed, so save the pragmatic model
    #         self.knowledge_database.insert(pragmatic)
    #
    # def deconstruct_time_sigma(self,
    #                            tier: int,
    #                            pragmatic: PragmaticMachineLearningModel,
    #                            relative: PragmaticMachineLearningModel,
    #                            block) -> tuple:
    #     decon = TSigmaDeconstructor()
    #     success = decon.deconstruct(pragmatic, relative, block)
    #     return success
    #
    # def deconstruct_sigma_zeta(self,
    #                            tier: int,
    #                            pragmatic: PragmaticMachineLearningModel,
    #                            relative: PragmaticMachineLearningModel,
    #                            block) -> tuple:
    #     decon = ZSigmaDeconstructor(tier,
    #                                 self.source,
    #                                 self.knowledge_database,
    #                                 self.settings,
    #                                 self.logger,
    #                                 self.reconstructor)
    #     success = decon.deconstruct(pragmatic, relative, block)
    #     return success
    #
    # def deconstruct_complete(self,
    #                          tier: int,
    #                          pragmatic: PragmaticMachineLearningModel,
    #                          relative: PragmaticMachineLearningModel,
    #                          block) -> tuple:
    #     decon = CompleteDeconstructor(tier,
    #                                   self.source,
    #                                   self.knowledge_database,
    #                                   self.settings,
    #                                   self.logger,
    #                                   self.reconstructor)
    #     success = decon.deconstruct(pragmatic, relative, block)
    #     return success

    # def deconstruct_time_zeta(self,
    #                           tier: int,
    #                           prag_model: PragmaticMachineLearningModel,
    #                           relative_model: PragmaticMachineLearningModel,
    #                           block):
    #     success = False
    #     self.knowledge_database.insert(prag_model)
    #
    #     if relative_model:
    #         # Get learnblock that trained relative model
    #         second_block = relative_model.trained_with(self.source)
    #
    #         # Get samples that have overlapping timestamp
    #         over_block = block.overlapping_rows(second_block, subset=["T"])
    #         if over_block.rows >= self.settings.learn_block_minimum:
    #
    #             # Create new metadata for a pragmatic model
    #             new_model = prag_model.fusion(
    #                 relative_model, self.next_model_counter(tier))
    #
    #             # Which models should be used for the reconstruction
    #             which_ml_models = new_model.sigma
    #
    #             # Get learningblock
    #             train_block = block.fusion(second_block)
    #
    #             # Start the reconstruction
    #             try:
    #                 recon_model = self.reconstructor.reconstruct(
    #                     tier, train_block, which_ml_models, new_model)
    #                 self.knowledge_database.replace(relative_model, recon_model)
    #                 success = True
    #             except (NoModelReconstructedError, NotEnoughFeaturesWarning):
    #                 pass
    #             except ValueError as error:
    #                 print(error.with_traceback())
    #
    #     if not success:
    #         raise DeconstructionFailed()












