import sys
from collections import namedtuple
from multiprocessing import Process

import krippendorff


class AbstractDeconstructor(Process):
    def __init__(self,
                 tier,
                 source,
                 knowledge,
                 settings,
                 logger,
                 reconstructor,
                 pragmatic,
                 relatives,
                 block,
                 changes,
                 val,
                 *args):
        super(Process, self).__init__(name=self.__class__.__name__)
        self.tier = tier
        self.source = source
        self.knowledge = knowledge
        self.settings = settings
        self.logger = logger
        self.reconstructor = reconstructor
        self.pragmatic = pragmatic
        self.relatives = relatives
        self.block = block
        self.changes = changes
        self.count_from = val
        self.event = args[0] if args else None

    def run(self):
        try:
            for relative in self.relatives:
                change = self.deconstruct(relative)
                self.changes.put(change)
            if self.event:
                self.event.set()
        except Exception as e:
            print(e)
            print(e.with_traceback())
            self.event.set()
            # parent = getppid()
            # signal.signal(signal.SIGTERM, parent)
            sys.exit()

    Disposal = namedtuple("Disposal", "inserted, deleted")

    def deconstruct(self, relative):
        raise NotImplementedError()

    def model_disposal(self, pragmatic, relative, block, relative_block):
        if self.settings.deconst_strategy == "conservative":
            # self.knowledge.remove(pragmatic)
            return self.Disposal(inserted=[], deleted=[pragmatic])
        elif self.settings.deconst_strategy == "integrative":
            # self.knowledge.insert(pragmatic)
            return self.Disposal(inserted=[pragmatic], deleted=[])
        elif self.settings.deconst_strategy == "oppurtunistic":
            if block.rows > relative_block.rows:
                # self.knowledge.remove(relative)
                # self.knowledge.insert(pragmatic)
                return self.Disposal(inserted=[pragmatic], deleted=[relative])
            else:
                # self.knowledge.remove(pragmatic)
                # self.knowledge.insert(relative)
                return self.Disposal(inserted=[relative], deleted=[pragmatic])

    def calculate_reliability(self, predicts_a, predicts_b):
        # TODO (dmt): Krippendorff.alpha needs numpy float64 values to work
        # correct or an TypeError exception is thrown.
        predicts_b = [int(i) for i in predicts_b]
        predicts_a = [int(i) for i in predicts_a]
        predictions = [predicts_b, predicts_a]
        return krippendorff.alpha(predictions, level_of_measurement="nominal")

    def correct_reconstructed_counter(self, model):
        model.identifier = self.count_from.value
        with self.count_from.get_lock():
            self.count_from.value += 1
        return model
