from time import clock_gettime_ns, CLOCK_REALTIME
from typing import List
from dataclasses import dataclass
from contextlib import suppress

from cml.shared.logger import DeconstInfo
from cml.domain.deconstruction.abstract import AbstractDeconstructor
from cml.shared.errors import (
    NotEnoughFeaturesWarning,
    NoModelReconstructedError)

__all__ = (
    "CompleteChange",
    "CompleteDeconstructor"
)


@dataclass
class CompleteChange:
    time_stamp: int
    relative: object
    status: bool
    logs: DeconstInfo
    inserted: List[object]
    deleted: List[object]

    @staticmethod
    def success(log, *, inserted, deleted):
        finished_time = clock_gettime_ns(CLOCK_REALTIME)
        return CompleteChange(finished_time, log.relative, True, log, inserted,
                              deleted)

    @staticmethod
    def failure(log, *, inserted, deleted):
        finished_time = clock_gettime_ns(CLOCK_REALTIME)
        return CompleteChange(finished_time, log.relative, False, log, inserted,
                              deleted)


class CompleteDeconstructor(AbstractDeconstructor):
    RELATION = ("complete", )
    MIN_N_FEATURE = 2

    def deconstruct(self, relative):
        log = self.arrange_logs(relative)

        relative_block = relative.trained_with(self.source, self.knowledge)
        if self.feature_constraint(relative):
            meta = self.model_fusion(relative)
            learnblock = self.block.same_features_fusion(relative_block)
            return self.reconstructing(meta, learnblock, relative, log)

        elif self.reliability_constraint(relative_block):
            meta = self.model_fusion(relative)
            learnblock = self.build_block_from_diff_features(relative_block)
            return self.reconstructing(meta, learnblock, self.pragmatic, log)

        else:
            change = self.model_disposal(self.pragmatic, relative,
                                         self.block, relative_block)
            return CompleteChange.failure(log,
                                          inserted=[change.inserted],
                                          deleted=[change.deleted])

    def arrange_logs(self, relative):
        log = DeconstInfo()
        log.relationship = self.RELATION
        log.relative = relative
        log.pragmatic = self.pragmatic
        return log

    def feature_constraint(self, relative):
        first_block = self.pragmatic.trained_with(self.source, self.knowledge)
        second_block = relative.trained_with(self.source, self.knowledge)
        identicial_features = len(set(first_block.columns()).intersection(
            set(second_block.columns())))
        return identicial_features >= self.MIN_N_FEATURE

    def model_fusion(self, relative):
        model_counter = self.knowledge.model_counter()
        return self.pragmatic.fusion(relative, self.tier,
                                     model_counter(self.tier))

    def reconstructing(self, meta, learnblock, relative, log):
        try:
            reconstructed = self.reconstructor.reconstruct_serial(
                self.tier, learnblock, meta.sigma, meta)
            corrected_recon = self.correct_reconstructed_counter(reconstructed)
            return CompleteChange.success(log,
                                          inserted=[corrected_recon],
                                          deleted=[relative])
            # self.knowledge.remove(relative)
            # self.knowledge.insert(recon_model)
        except (NoModelReconstructedError, NotEnoughFeaturesWarning):
            return self.model_differentiation(self.tier, learnblock, relative,
                                              log)

    def model_differentiation(self, tier, block, relative_model, log):
        time_column = block.get_column("T")
        density = self.source.estimate_density(time_column)
        self.source.remove_time_dense_relatives(block, density)
        clusters = self.source.cluster(block, density)

        inserted = []
        for time_values in clusters:
            learnblock = block.new_block_from(time_values, time_column=True)
            with suppress((NoModelReconstructedError, NotEnoughFeaturesWarning)):
                reconstructed = self.reconstructor.reconstruct_serial(tier, learnblock)
                corrected_recon = self.correct_reconstructed_counter(
                    reconstructed)
                inserted.append(corrected_recon)
                # self.knowledge.insert(reconstructed_model)

        if not inserted:
            return CompleteChange.success(log, inserted=inserted, deleted=[])
        else:
            deleted = self.knowledge.remove_dependent_models(relative_model)
            return CompleteChange.failure(log, inserted=[], deleted=deleted)
            # self.knowledge_database.remove(relative_model)

    def reliability_constraint(self, relative_block):
        overlapped = self.block.get_overlapping(relative_block)
        reliability = self.calculate_reliability(
            overlapped.get_column("Z"),
            overlapped.get_column("Z_other"))
        return reliability >= self.settings.min_reliability

    def build_block_from_diff_features(self, relative_block):
        features = set(self.block.n_features).difference(
            set(relative_block.n_features))
        features.add("T")
        return self.block.get_overlapping(relative_block,
                                          drop_not=tuple(features))

