from time import clock_gettime_ns, CLOCK_REALTIME
from dataclasses import dataclass

from cml.domain.deconstruction.abstract import AbstractDeconstructor
from cml.shared.logger import DeconstInfo

__all__ = (
    "TSigmaChange",
    "TSigmaDeconstructor"
)


@dataclass
class TSigmaChange:
    time_stamp: int
    relative: object
    status: bool
    logs: DeconstInfo
    new_learnblock: object = None
    inserted: list = None
    deleted: list = None

    @staticmethod
    def success(block, log):
        finished_time = clock_gettime_ns(CLOCK_REALTIME)
        return TSigmaChange(finished_time, log.relative, True, log, block,
                            inserted=[], deleted=[])

    @staticmethod
    def failure(log):
        finished_time = clock_gettime_ns(CLOCK_REALTIME)
        return TSigmaChange(finished_time, log.relative, False, log,
                            inserted=[], deleted=[])


class TSigmaDeconstructor(AbstractDeconstructor):
    RELATION = ("Sigma", "T")

    def deconstruct(self, relative):
        log = self.arrange_logs(relative)
        if self.tier_constraint():
            overlapping = self.identify_overlapping_block(relative)
            if self.row_and_reliability_contraints(overlapping):
                block = self.prepare_higher_block(relative, overlapping)
                return TSigmaChange.success(block, log)
                # TS_QUEUE.append((self.tier + 1, overblock)) Move this to main
        return TSigmaChange.failure(log)

    def arrange_logs(self, relative):
        log = DeconstInfo()
        log.relationship = self.RELATION
        log.relative = relative
        log.pragmatic = self.pragmatic
        return log

    def tier_constraint(self):
        return self.pragmatic.tier < self.settings.highest_tier - 1

    def identify_overlapping_block(self, relative):
        relative_block = relative.trained_with(self.source, self.knowledge)
        overlapped_block = self.block.get_overlapping(relative_block)
        return overlapped_block

    def row_and_reliability_contraints(self, block):
        return self.row_constraint(block) and self.reliability_constraint(block)

    def row_constraint(self, block):
        return len(block) >= self.settings.learn_block_minimum

    def reliability_constraint(self, block):
        alpha = self.calculate_reliability(
            block.get_column("Z"),
            block.get_column("Z_other"))
        alpha_systematic = alpha < 0
        alpha_weak_reliability = 0 <= alpha < self.settings.min_reliability
        return (self.settings.allow_weak_reliability and alpha_weak_reliability) or alpha_systematic

    def prepare_higher_block(self, relative, overlapping):
        overblock = self.source.new_learnblock(
            values=list(zip(
                overlapping.get_column("Z"),
                overlapping.get_column("Z_other"),
                overlapping.get_column("T"),
                ("\"\"" for _ in range(overlapping.n_rows())),
                ("\"\"" for _ in range(overlapping.n_rows())))),
            columns=(self.pragmatic.uid, relative.uid, "T", "Sigma", "Z"),
            index=[i for i in range(overlapping.n_rows())],
            origin=(self.pragmatic.uid, relative.uid))
        return overblock
