from time import clock_gettime_ns, CLOCK_REALTIME
from dataclasses import dataclass

from cml.shared.logger import DeconstInfo
from cml.domain.deconstruction.abstract import AbstractDeconstructor
from cml.shared.errors import (
    NotEnoughFeaturesWarning,
    NoModelReconstructedError)

__all__ = (
    "ZSigmaChange",
    "ZSigmaDeconstructor"
)


@dataclass
class ZSigmaChange:
    time_stamp: int
    relative: object
    status: bool
    logs: DeconstInfo
    inserted: list
    deleted: list

    @staticmethod
    def success(log, *, inserted, deleted):
        finished_time = clock_gettime_ns(CLOCK_REALTIME)
        return ZSigmaChange(finished_time, log.relative, True, log,
                            inserted, deleted)

    @staticmethod
    def failure(log, *, inserted, deleted):
        finished_time = clock_gettime_ns(CLOCK_REALTIME)
        return ZSigmaChange(finished_time, log.relative, False, log,
                            inserted, deleted)


class ZSigmaDeconstructor(AbstractDeconstructor):
    RELATION = ("Sigma", "Z")
    MIN_N_FEATURE = 2

    def deconstruct(self, relative):
        log = self.arrange_logs(relative)
        if self.time_constraint(relative):
            overlapping = self.identify_overlapping_features(relative)
            if self.column_constraint(overlapping):
                meta_model = self.model_fusion(relative)
                return self.reconstructing(overlapping, meta_model, relative, log)
                    # self.knowledge.replace(relative, recon_m) Move this to main

        return ZSigmaChange.failure(log, inserted=[], deleted=[])

    def arrange_logs(self, relative):
        log = DeconstInfo()
        log.relationship = self.RELATION
        log.relative = relative
        log.pragmatic = self.pragmatic
        return log

    def time_constraint(self, relative):
        if self.settings.deconst_max_distance_t == 0:
            return self.overlapping_time(self.pragmatic, relative)

        elif 0 < self.settings.deconst_max_distance_t < 1:
            return self.enclosing_time(self.pragmatic, relative)

        elif self.settings.deconst_max_distance_t == 1:
            return True

    def overlapping_time(self, prag_model, relative_model):
        return relative_model.min_timestamp >= prag_model.max_timestamp or \
            prag_model.min_timestamp >= relative_model.max_timestamp

    def enclosing_time(self, prag_model, relative_model):
        m_dash_max_time = max(prag_model.max_timestamp,
                              relative_model.max_timestamp)
        m_dash_min_time = min(prag_model.min_timestamp,
                              relative_model.min_timestamp)
        relative_m_dash_time = self.settings.deconst_max_distance_t*(
                m_dash_max_time-m_dash_min_time)

        return (relative_model.min_timestamp - prag_model.max_timestamp) < \
               relative_m_dash_time or (prag_model.min_timestamp -
                                        relative_model.max_timestamp) < relative_m_dash_time

    def identify_overlapping_features(self, relative):
        relative_block = relative.trained_with(self.source, self.knowledge)
        overlapped_block = self.block.same_features_fusion(relative_block)
        return overlapped_block

    def column_constraint(self, block):
        return block.n_columns(effective=True) >= self.MIN_N_FEATURE

    def model_fusion(self, relative):
        model_counter = self.knowledge.model_counter()
        meta_model = self.pragmatic.fusion(
            relative,
            self.tier,
            model_counter(self.tier))
        return meta_model

    def reconstructing(self, block, meta_model, relative, log):
        try:
            ml_models = meta_model.sigma
            reconstructed = self.reconstructor.reconstruct_serial(
                self.tier, block, ml_models, meta_model)
            corrected_recon = self.correct_reconstructed_counter(reconstructed)
            return ZSigmaChange.success(log,
                                        inserted=[corrected_recon],
                                        deleted=[relative])

        except (NoModelReconstructedError, NotEnoughFeaturesWarning):
            relative_block = relative.trained_with(self.source, self.knowledge)
            change = self.model_disposal(self.pragmatic, relative,
                                         block, relative_block)
            return ZSigmaChange.failure(log,
                                        inserted=change.inserted,
                                        deleted=change.deleted)

