from random import sample, seed, shuffle
from collections import defaultdict
from itertools import count
from functools import partial
from multiprocessing import Pool, cpu_count
from typing import Union, List, Tuple, Dict

from krippendorff import alpha

from cml.shared.logger import ReconstLogger
from cml.domain.data_source import DataSource
from cml.shared.settings import ReconstructionSettings
from cml.shared.errors import (
    NoModelReconstructedError,
    NotEnoughFeaturesWarning)

__all__ = (
    "Reconstructor",
)


class Metadata:
    def __init__(self,
                 knowledge_domain,
                 knowledge_tier,
                 identifier,
                 pre_image,
                 pre_image_features,
                 pre_image_labels,
                 t_min,
                 t_max,
                 sigma,
                 zeta):
        """
        Sigma = ANN, RF
        Zeta: C.2.Kme02
        """
        assert isinstance(sigma, tuple)

        self.knowledge_domain = knowledge_domain
        self.knowledge_tier = knowledge_tier
        self.identifier = identifier
        self.pre_image = pre_image
        self.pre_image_features = pre_image_features
        self.pre_image_labels = pre_image_labels
        self.t_min = t_min
        self.t_max = t_max
        self.sigma = sigma
        self.zeta = zeta

    def __str__(self):
        return f"Knowledge domain: <{self.knowledge_domain}> \n" \
               f"Knowledge tier: <{self.knowledge_tier}> \n" \
               f"Identifier: <{self.identifier}> \n" \
               f"Pre image: <{self.pre_image}> \n" \
               f"Pre image labels: <{self.pre_image_labels}> \n" \
               f"Pre image features: <{self.pre_image_features}> \n" \
               f"T min: <{self.t_min}> \n" \
               f"T max: <{self.t_max}> \n" \
               f"Subjects: <{self.sigma}> \n" \
               f"Purposes: <{self.zeta}> \n"

    def __hash__(self):
        return hash(".".join([self.knowledge_domain,
                              str(self.knowledge_tier),
                              str(self.identifier)]))

    def __eq__(self, other):
        if not isinstance(other, Metadata):
            raise TypeError()

        return hash(self) == hash(other)


class PragmaticMachineLearningModel:

    def __init__(self,
                 meta: Metadata,
                 model,
                 learnblock):
        self.meta = meta
        self.model = model
        self.origin = learnblock.origin
        self.learnblock = learnblock if self.tier > 1 else None
        assert isinstance(self.origin, tuple)

    def __str__(self) -> str:
        return self.uid

    def __repr__(self) -> str:
        return self.uid

    def __hash__(self) -> int:
        return hash(self.meta)

    def __eq__(self,
               other: Union['PragmaticMachineLearningModel', str]) -> bool:
        if isinstance(other, PragmaticMachineLearningModel):
            return hash(self.meta) == hash(other.meta)

        if isinstance(other, str):
            return hash(self) == hash(other)

        raise NotImplementedError()

    @property
    def pre_image_features(self) -> List[str]:
        return self.meta.pre_image_features

    @property
    def tier(self) -> int:
        return self.meta.knowledge_tier

    @property
    def min_timestamp(self) -> int:
        return self.meta.t_min

    @property
    def max_timestamp(self) -> int:
        return self.meta.t_max

    @property
    def pre_image(self) -> List[int]:
        return self.meta.pre_image

    @property
    def pre_image_labels(self) -> List[Union[int, str]]:
        return self.meta.pre_image_labels

    @property
    def subject(self) -> Tuple[str]:
        return self.meta.sigma

    @property
    def purpose(self) -> Tuple[str]:
        return self.meta.zeta

    @property
    def uid(self) -> str:
        return ".".join([self.meta.knowledge_domain,
                         str(self.meta.knowledge_tier),
                         str(self.meta.identifier)])

    @property
    def identifier(self):
        return self.meta.identifier

    @identifier.setter
    def identifier(self, value):
        self.meta.identifier = value

    @property
    def counter(self):
        return self.meta.identifier

    @property
    def accuracy(self):
        return self.model.accuracy

    @property
    def reliability(self):
        return self.model.reliability

    def fusion(self,
               model: 'PragmaticMachineLearningModel',
               knowledge_tier: int,
               new_identifier: int) -> Metadata:

        return Metadata(self.meta.knowledge_domain,
                        knowledge_tier,
                        new_identifier,
                        None,
                        None,
                        None,
                        min(self.meta.t_min, model.min_timestamp),
                        max(self.meta.t_max, model.max_timestamp),
                        tuple(set(self.subject + model.subject)),
                        tuple(set(self.meta.zeta + model.purpose)))

    def trained_with(self, source: DataSource, knowledge):
        if "source" in self.origin:
            if self.origin[1] == source.n_source:
                true_origin = source
            else:
                true_origin = knowledge.sources[self.origin[1]]

            block = true_origin.get_block(self.pre_image, columns=self.pre_image_features)
            block.set_column("Z", self.pre_image_labels)
            block.subject = self.subject
            return block

        else:
            return self.learnblock


class SubprocessDTO:
    def __init__(self, count ,model, learnblock, tier, abv, settings):
        self.count = count
        self.model = model
        self.learnblock = learnblock
        self.tier = tier
        self.abv = abv
        self.settings = settings


class Reconstructor:
    CONCEPTUAL_KNOWLEDGE_ABBREVIATION = "C"
    PROCEDURAL_KNOWLEDGE_ABBREVIATION = "P"

    def __init__(self,
                 settings: ReconstructionSettings,
                 ml_models: List,
                 mode: str):
        self.mode = mode
        self.settings = settings
        self.ml_models = ml_models
        self.__reconstruction = self._init_reconstruction(mode)

        self.next_valid_tier_count = None
        self.logger = None

    def inject_counter(self, counter):
        self.next_valid_tier_count = counter

    def _init_reconstruction(self, mode: str):
        if mode == "conceptual":
            return partial(self._reconstruct_conceptual, krippen="nominal")
        # else:
        #    return partial(self._reconstruct_procedural, krippen="ratio")

    def reconstruct_parallel(self, tier, learnblock, models=None, meta=None):
        from cml.domain.reconstruction_parallel import reconstruct_in_parallel
        from cml.domain.reconstruction_parallel import set_valid_counter

        if not models:
            args = []
            for count, model in enumerate(self.ml_models, 1):
                args.append(
                    SubprocessDTO(
                        count, model, learnblock, tier,
                        self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                        self.settings))

            set_valid_counter(len(self.ml_models))
            pool = Pool(cpu_count()-2)
            reliability_model = pool.map(reconstruct_in_parallel, args)
            pool.close()
            pool.join()
            reliabilities_to_model = defaultdict(list)
            for r, m in reliability_model:
                if r and m:
                    reliabilities_to_model[r].append(m)

            if reliabilities_to_model.keys():
                winner = determine_winner(reliabilities_to_model)
                winner.identifier = self.next_valid_tier_count(tier)
                return winner
            raise NoModelReconstructedError()

    def reconstruct_serial(self,
                    tier: int,
                    learnblock,
                    which_models: List = None,
                    meta: Metadata = None) -> PragmaticMachineLearningModel:

        # Check if learnblock has enough features
        if not (learnblock.n_columns(effective=True) >= 2):
            raise NotEnoughFeaturesWarning()

        # Specify the models which should be trained
        if not which_models:
            which_models = [m.subject for m in self.ml_models]

        # Start the reconstruction
        reliabilities_to_model = self.__reconstruction(tier,
                                                       learnblock,
                                                       which_models,
                                                       meta=meta)
        # Determine the best pragmatic machine learning model
        if reliabilities_to_model.keys():
            winner = determine_winner(reliabilities_to_model)
            winner.identifier = self.next_valid_tier_count(tier)
            return winner
        raise NoModelReconstructedError()

    def _reconstruct_conceptual(self,
                                tier: int,
                                learnblock,
                                which_models: List,
                                krippen: str = None,
                                meta: Metadata = None):
        reliability_to_model = defaultdict(list)
        for count, model in enumerate(self.ml_models, 1):
            if model.subject not in which_models: continue

            train_block, eval_block = split(learnblock, self.settings.reliability_sample)
            data = train_block.as_numpy_array()
            labels = [label for label in train_block.get_column("Z")]
            trained_model = model.train(data, labels)

            if valid_reconstructed(trained_model,
                                   "conceptual",
                                   self.settings.min_test_accuracy,
                                   self.settings.max_test_error_avg,
                                   self.settings.max_test_error_max):
                reliability = calc_reliability(trained_model, learnblock, krippen)
                if reliability >= self.settings.min_reliability:
                    trained_model.reliability = reliability
                    if not meta:
                        uid = ".".join([self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                                        str(tier),
                                        learnblock.subject])
                        meta_data = Metadata(self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                                            tier,
                                            count,
                                            learnblock.indices(),
                                            learnblock.columns(),
                                            learnblock.get_column("Z"),
                                            learnblock.min_timestamp(),
                                            learnblock.max_timestamp(),
                                            (model.subject, ),
                                            (uid, ))
                    else:
                        meta_data = meta
                        meta_data.pre_image_features = learnblock.columns()
                        meta_data.pre_image_labels = learnblock.get_column("Z")
                        meta_data.pre_image = learnblock.indices()

                    reliability_to_model[reliability].append(
                        PragmaticMachineLearningModel(
                            meta_data, trained_model, learnblock))
        return reliability_to_model

    # def _reconstruct_procedural(self,
    #                             tier: int,
    #                             learnblock,
    #                             which_models: List,
    #                             krippen: str = None,
    #                             meta: Metadata = None) -> Dict[float, List]:
    #     reliability_to_model = defaultdict(list)
    #     for model in self.ml_models:
    #         if model.subject not in which_models: continue
    #
    #         # train model
    #         train_block, eval_block = self.split(learnblock)
    #         trained_model = model.train(
    #             train_block.as_numpy_array(),
    #             [i for i in train_block.get_column("Z")])
    #
    #         # check contraints
    #         if valid_reconstructed(trained_model, "procedural"):
    #             reliability = self.calc_reliability(
    #                 trained_model, learnblock, krippen)
    #             if reliability >= self.settings.min_reliability:
    #                 if not meta:
    #                     prag_meta_data = Metadata(
    #                         "P",
    #                         tier,
    #                         next(self.next_model_counter),
    #                         learnblock.indices(),
    #                         learnblock.columns(),
    #                         learnblock.get_column("Z"),
    #                         learnblock.min_timestamp(),
    #                         learnblock.max_timestamp(),
    #                         (model.subject, ),
    #                         (".".join(["C", tier, learnblock.subject]), )
    #                     )
    #                 else:
    #                     meta.pre_image_features = learnblock.columns()
    #                     meta.pre_image_labels = learnblock.get_column("Z")
    #                     meta.pre_image = learnblock.indices()
    #
    #                 reliability_to_model[reliability].append(
    #                     PragmaticMachineLearningModel(prag_meta_data,
    #                                                   trained_model,
    #                                                   learnblock))
    #     return reliability_to_model


def split(learnblock, reliability_sample) -> Tuple:
    seed(7)
    indices = learnblock.indices()
    eval_size = int(learnblock.n_rows() * reliability_sample)
    # eval_idx = sample(indices, eval_size)
    shuffle(indices)
    eval_idx, train_idx = indices[:eval_size], indices[eval_size+1:]

    # train_idx = list(set(indices).difference(set(eval_idx)))
    return learnblock.new_block_from(train_idx, index=True), \
        learnblock.new_block_from(eval_idx, index=True)


def calc_reliability(trained_model, eval_block, metric: str) -> float:
    y_pre = trained_model.predict(eval_block.as_numpy_array())
    y_true = [i for i in eval_block.get_column("Z")]
    reliability_data = [y_pre, y_true]
    return alpha(reliability_data, level_of_measurement=metric)


def valid_reconstructed(model,
                        knowledge_domain: str,
                        min_test_accuracy,
                        max_test_error_avg,
                        max_test_error_max) -> bool:
    if knowledge_domain == "conceptual":
        return model.accuracy >= min_test_accuracy
    # else:
    #    return model.mean_error <= max_test_error_avg and \
    #        model.max_error <= max_test_error_max


def determine_winner(
        reliability_to_model: dict) -> PragmaticMachineLearningModel:
    sorted_reliabilities = sorted(reliability_to_model.keys(), reverse=False)
    biggest_reliabilities = reliability_to_model[sorted_reliabilities.pop()]
    winner = None
    min_domain = float("inf")
    for model in biggest_reliabilities:
        if len(model.pre_image_features) < min_domain:
            min_domain = len(model.pre_image_features)
            winner = model

    return winner
