from os import mkdir
from os.path import expanduser, join
from datetime import datetime
from logging import getLogger, FileHandler, StreamHandler
from collections import namedtuple
from contextlib import suppress
import multiprocessing

from cml.shared.logger import TierBanner, ConstLogger, ReconstLogger
from cml.shared.parameter import PROTOCOL_LEVEL
from cml.usecases.usecase import Usecase
from cml.domain.knowledge import TS_QUEUE
from cml.shared.errors import (
    NoModelReconstructedError,
    NotReducableLearnblockWarning,
    ModiError)


Components = namedtuple(
    "Components",
    "source constructor selector reconstructor deconstructor")


class SourceWrapper:
    def __init__(self, source):
        self.source = source

    def prepare(self, logger, old_source_size, start_index, end_index):
        self.source.logger = logger
        self.source.n_source = old_source_size
        self.source.reindex(start_index, end_index)
        self.source.remap_feature_names()

    @TierBanner
    def learnblock(self):
        first_tier = 1
        for lb in self.source.learnblocks:
            yield first_tier, lb

        TS_QUEUE.reverse()
        for tier, lb in TS_QUEUE:
            if tier and lb:
                yield tier, lb


class ConstructorWrapper:
    def __init__(self, con):
        self.con = con
        self.act = None

    def prepare(self, logger, parallel=False):
        self.con.logger = logger
        if parallel:
            self.act = self.con.construct_parallel
        else:
            self.act = self.con.construct_serial

    @ConstLogger
    def construct(self, learnblock):
        for log_info, labeled in self.act(learnblock):
            yield log_info, labeled


class FeatureSelectorWrapper:
    def __init__(self, selector):
        self.selector = selector

    def prepare(self, logger):
        self.selector.logger = logger

    def select(self, labeled):
        return self.selector.select(labeled)


class ReconstructorWrapper:
    def __init__(self, recon):
        self.recon = recon
        self.act = None

    def prepare(self, logger, model_counter, parallel=False):
        self.recon.logger = logger
        self.recon.inject_counter(model_counter)

        if parallel:
            self.act = self.recon.reconstruct_parallel
        else:
            self.act = self.recon.reconstruct_serial

    @ReconstLogger
    def reconstruct(self, tier, reduced):
        return self.act(tier, reduced)


class DeconstructorWrapper:
    def __init__(self, decon):
        self.decon = decon
        self.act = None

    def prepare(self, logger, recon, parallel=False):
        self.decon.logger = logger
        self.decon.reconstructor = recon
        if parallel:
            self.act = self.decon.deconstruct_parallel
        else:
            self.act = self.decon.deconstruct_serial

    def deconstruct(self, tier, pragmatic, reduced):
        self.act(tier, pragmatic, reduced)

    def finish(self, source):
        self.decon.knowledge_database.insert_source(source)


class KnowledgeSearchUsecase(Usecase):
    FIRST_TIER = 1

    def __init__(self):
        super().__init__()
        self.logger = None

    def process(self, request):
        self.valid_constraints(request)
        wrappers = self.prepare_wrappers(request)
        self.search(wrappers)
        return self.finish_wrappers(wrappers)

    def valid_constraints(self, req):
        if not self.valid_modi(req):
            raise ModiError()

    def prepare_wrappers(self, req):
        logger = self.prepare_logger(req)
        self.logger = logger

        # Prepare data source
        n_source = len(req.deconstructor.knowledge_database.sources)
        start_index = req.deconstructor.knowledge_database.last_index_source()
        end_index = start_index+len(req.deconstructor.source)
        source = SourceWrapper(req.deconstructor.source)
        source.prepare(logger, n_source, start_index, end_index)

        # Prepare constructor
        con = ConstructorWrapper(req.constructor)
        con.prepare(logger, req.parallel)

        # Prepare selector
        selector = FeatureSelectorWrapper(req.feature_selector)
        selector.prepare(logger)

        # Prepare reconstructor
        recon = ReconstructorWrapper(req.reconstructor)
        recon.prepare(logger, req.deconstructor.next_model_counter, req.parallel)

        # Prepare deconstructor
        decon = DeconstructorWrapper(req.deconstructor)
        decon.prepare(logger, req.reconstructor, req.parallel)

        return source, con, selector, recon, decon

    def prepare_logger(self, req):
        log_dir = create_log_dir()
        iteration_log_file = join(log_dir, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")

        # Remove loggers from other sessions
        if iteration_logger.handlers: iteration_logger.handlers.clear()
        iteration_logger.addHandler(iteration_handler)

        if req.stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)
        return iteration_logger

    def search(self, wrappers):
        with suppress(NotReducableLearnblockWarning, NoModelReconstructedError):
            for _, lb in wrappers[0].learnblock():
                for _, labeled in wrappers[1].construct(lb):
                    reduced = wrappers[2].select(labeled)
                    pragmatic = wrappers[3].reconstruct(self.FIRST_TIER, reduced)
                    wrappers[4].deconstruct(self.FIRST_TIER, pragmatic, reduced)

    def finish_wrappers(self, wrappers):
        wrappers[4].finish(wrappers[0].source)

        for child in multiprocessing.active_children():
            child.kill()

        return wrappers[4].decon.knowledge_database

    def valid_modi(self, request):
        return request.constructor.mode == request.reconstructor.mode == \
               request.deconstructor.mode

    def close_background_processes(self):
        for child in multiprocessing.active_children():
            child.kill()


def create_log_dir():
    home_folder_path = expanduser("~")
    cml_folder = ".cml"
    current_time = datetime.now().strftime("%d%m%Y%H%M%S")
    log_dir = join(home_folder_path, cml_folder, current_time)
    mkdir(log_dir)
    return log_dir

